-- common.e
-- library of common reusable functions
-- v1.07
-- 2010-10-04
-- history:
-- v1.0   2010-07-26
-- v1.01
--   changed french terms to english ones
--   saveCSV uses quotes around strings
-- v1.02
--   changed id from atom to object in findID
-- v1.03  2010-07-28
--   added dateStamp function
-- v1.04  2010-07-29
--   added min and max functions
-- v1.05  2010-08-04
--   added  function appendIfNew
-- v1.06  2010-09-29
--   added function utf8_to_ascii
-- v1.07  2010-10-04
--   new INI files management
-- v1.08  2010-11-05
--   adde searchINI

include get.e
include md5.e
include msgbox.e
include file.e

--include tk_trim.e
--with trace

global integer f_debug,    -- debug file
       log_status, log_level

global constant   -- log status
  ON = 1,
  OFF = 0

global constant   -- log lvl
  LOG_ERRORS=1,
  LOG_WARNINGS=2,
  LOG_ALL=3
  
-- request result
global constant SUCCESS = 1,
         FAIL = 0

object void    -- used when return value doesn't care
         
------------------------------------------------------------------------------

global function dateStamp(sequence msg)
-- prefixes the message by date and time for logging
  sequence cur_date

  cur_date = date()
  return sprintf("%d-%02d-%02d %02d:%02d:%02d -> %s\n", {
                             (cur_date[1] + 1900), cur_date[2], cur_date[3],
                              cur_date[4], cur_date[5], cur_date[6], msg
                            })
end function

------------------------------------------------------------------------------

global procedure logMsg(sequence msg, integer lvl)
-- records logs according to the log status and log level
-- global variable log_status determins if logging is enabled
  if log_status then
    if ((log_level=LOG_ERRORS) and (lvl=1)) or
       ((log_level=LOG_WARNINGS) and (lvl>=2)) or
       (log_level=LOG_ALL)
    then
      puts(f_debug, dateStamp(msg))
      flush(f_debug)
    end if
  end if
end procedure

------------------------------------------------------------------------------

global procedure warnError(sequence msg, atom quit)
-- logs an error and displays a warning
-- if quit is set then abort
  logMsg(msg, LOG_ERRORS)
  if platform() = WIN32 then
    void = message_box(msg, "Erreur", MB_OK+MB_ICONERROR)
  end if
  if quit then
    abort(1)
  end if
end procedure

------------------------------------------------------------------------------

global function leftTrim(sequence s)
-- remove space chars from the beginning of a string
  integer l

  l = length(s)
  if l=0 then return "" end if
  for i=1 to l do
    if s[i] != 32 then
      return s[i..l]
    end if
  end for
  return ""
end function

------------------------------------------------------------------------------

global function rightTrim(sequence s)
-- remove space chars from the end of a string
  integer l

  l = length(s)
  if l=0 then return "" end if
  for i=l to 1 by -1 do
    if s[i] != 32 then
      return s[1..i]
    end if
  end for
  return ""
end function

------------------------------------------------------------------------------

global function trim(sequence s)
-- remove space chars from both ends of a string
  return leftTrim(rightTrim(s))
end function

------------------------------------------------------------------------------

global function rightPad(sequence s, atom len)
--Return a string right-padded to length len

  if length(s) >= len then
    return s
  else
    return s & repeat(' ', len-length(s))
  end if
end function

------------------------------------------------------------------------------

global function leftPad(sequence s, atom len)
--Return a string left-padded to length len

  if length(s) >= len then
    return s
  else
    return repeat(' ', len-length(s)) & s
  end if
end function

------------------------------------------------------------------------------

global function right(sequence s, integer n)
-- returns the n characters at the right of string s
  integer l

  l = length(s)
  if (l=0) or (n=0) then return "" end if
  if n >= l then return s end if
  return s[l-n+1..l]
end function

------------------------------------------------------------------------------

global type string(object x)
-- checks if a sequence is a string
  atom is_string
  if sequence(x) then
    is_string = 1
    for i = 1 to length(x) do
      if sequence(x[i]) then
        is_string = 0
      else
        if (x[i] < ' ') and (x[i] != 9) and (x[i] != 13) and (x[i] != 10) then is_string = 0 end if
        if x[i]> 254 then is_string = 0 end if
      end if
    end for
  else
    is_string = 0
  end if
  return is_string
end type

------------------------------------------------------------------------------

global function isNumber(sequence x)
-- checks if a string represents a number
  for i = 1 to length(x) do
    if sequence(x[i]) then
      return 0
    else
      if ((x[i] < '0') or (x[i] > '9')) and (x[i] != '-') and (x[i] != '+') then return 0 end if
    end if
  end for
  return 1
end function

------------------------------------------------------------------------------

global function max(sequence lst)
-- returns the highest value from the items of the list lst
-- returns a sequence with first item equal to SUCCESS if successful
--   and second item equal to the highest value
-- if unsuccessful, first item equals FAIL
  atom maxi
  
  for i = 1 to length(lst) do
    if sequence(lst[i]) then
      return {FAIL, 0}
    elsif i=1 then
      maxi = lst[i]
    elsif lst[i] > maxi then
      maxi = lst[i]
    end if
  end for
  return {SUCCESS, maxi}
end function

------------------------------------------------------------------------------

global function min(sequence lst)
-- returns the lowest value from the items of the list lst
-- returns a sequence with first item equal to SUCCESS if successful
--   and second item equal to the lowest value
-- if unsuccessful, first item equals FAIL
  atom mini

  for i = 1 to length(lst) do
    if sequence(lst[i]) then
      return {FAIL, 0}
    elsif i=1 then
      mini = lst[i]
    elsif lst[i] < mini then
      mini = lst[i]
    end if
  end for
  return {SUCCESS, mini}
end function

------------------------------------------------------------------------------

global function val(sequence str)
-- returns numerical value of string str
-- decimal separator may be a comma
-- displays a warning on error
  integer n
  sequence s

  if length(str) = 0 then
    return 0
  end if
  n = find(',', str)
  if n>0 then str[n] = '.' end if
  s = value(str)
  if s[1] = GET_SUCCESS then
    return s[2]
  else
    warnError(sprintf("Erreur de conversion: %s", {str}), 1)
  end if
end function

------------------------------------------------------------------------------

global function dos2unix(sequence s)
-- converts DOS end of lines to Unix ones
  sequence t
  integer l

  t = {}
  l = match({13,10}, s)
  while l > 0 do
    t = t & s[1..l-1]
    s = s[l+1..length(s)]
    l = match({13,10}, s)
  end while
  t = t & s
  return t
end function

------------------------------------------------------------------------------

global function atom2string(atom val, sequence fmt)
-- converts numerical value val according to format fmt
-- replaces decimal separator by a comma (for non english speaking countries)
  integer n
  sequence s

  s = sprintf(fmt, val)
  n = find('.', s)
  if n>0 then s[n] = ',' end if
  return s
end function

------------------------------------------------------------------------------

global procedure analyzeSequence(object s, sequence variable, integer output)
-- prints a sequence structure in a human readable way
  if integer(s) then
    printf(output, "%s = %d\n", {variable, s})
  elsif atom(s) then
    if (s >= 0) and (s = floor(s)) then
      printf(output, "%s = %.0f\n", {variable, s})
    elsif (s < 0) and (s = floor(s+1)) then
      printf(output, "%s = %.0f\n", {variable, s})
    else
      printf(output, "%s = %f\n", {variable, s})
    end if
  elsif string(s) then
    printf(output, "%s = '%s'\n", {variable, s})
  else
    for i=1 to length(s) do
      analyzeSequence(s[i], variable&sprintf("[%d]", i), output)
    end for
  end if
  if output=f_debug then flush(f_debug) end if
end procedure

------------------------------------------------------------------------------

global function showPrintable(object s)
-- returns only printable characters of a sequence
-- non printable characters are replace by a dot
  sequence res
  
  res = ""
  for i = 1 to length(s) do
    if integer(s[i]) then
      if (s[i] > 31) and (s[i] < 255) then
        res &= s[i]
      elsif s[i] = 9 then
        res &= "\\t"
      elsif s[i] = 13 then
        res &= "\\r"
      elsif s[i] = 10 then
        res &= "\\n"
      else
        res &= "."
      end if
    else
      res &= "."
    end if
  end for
  return res
end function

------------------------------------------------------------------------------

global function object2string(object s)
-- converts an object in a sequence
-- sub-sequences appear between curly brackets as with print()
-- sub-sequence items are separated by commas as with print()
  sequence res

  if integer(s) then
    res = sprintf("%d", s)
  elsif atom(s) then
    if (s >= 0) and (s = floor(s)) then
      res = sprintf("%.0f", s)
    elsif (s < 0) and (s = floor(s+1)) then
      res = sprintf("%.0f", s)
    else
      res = sprintf("%f", s)
    end if
  elsif string(s) then
    res = s
  else
    res = {123}  -- {
    for i=1 to length(s) do
      res &= object2string(s[i])&","
    end for
    res[$] = 125   -- }
  end if
  return res
end function

------------------------------------------------------------------------------

global function replaceString(sequence str, sequence old, sequence new)
-- replaces sub-sequence old by new in sequence str
  sequence s
  integer n, lg

  s = str
  lg = length(old)
  n = -1
  while n != 0 do
    n = match(old, s)
    if n > 1 then
      s = s[1..n-1]&new&s[n+lg..$]
    elsif n = 1 then
      s = new&s[n+lg..$]
    end if
  end while
  return s
end function

------------------------------------------------------------------------------

global procedure textReplace(sequence old_file, sequence new_file, sequence old, sequence new)
-- replaces sub-sequence old by new in all sequences from file old_file
-- result is written in file new_file
-- old_file is not affected
  sequence s
  integer f_in, f_tmp
  object line

  s = {}
  f_in = open(old_file, "r")
  f_tmp = open("temp.$$$", "w")
  if (f_in != -1) and (f_tmp != -1) then
    while 1 do
      line = gets(f_in)
      if atom(line) then
        exit
      end if
      puts(f_tmp, replaceString(line, old, new))
    end while
    close(f_in)
    close(f_tmp)
  end if
  system("copy temp.$$$ " & new_file, 2)
  system("del temp.$$$", 2)
end procedure

------------------------------------------------------------------------------

global function splitString(sequence str, atom sep)
-- splits str string in tokens according to separator sep
  integer slen
  sequence result, s
  integer tmpStr

  result = {}
  slen = length(str)
  s = {}
  tmpStr = 0      -- si 1, on est entre deux guillemets

  for i = 1 to slen do
    if (str[i] = '"') and (sep != '"') then
      if tmpStr = 0 then
        tmpStr = 1
      else
        tmpStr = 0
      end if
    else
      if tmpStr = 0 and str[i] = sep then
        result = append(result, trim(s))
        s = {}
 	  else
	    s = s & str[i]
	  end if
    end if
  end for
  result = append(result, trim(s))
  return result
end function

------------------------------------------------------------------------------

global function strtok(sequence str, atom sep)
-- splits str string in tokens according to separator sep
-- converts segments to atoms when needed
  sequence s

  s = splitString(str, sep)
  for i = 1 to length(s) do
    if isNumber(s[i]) then s[i] = val(s[i]) end if
  end for
  return s
end function

------------------------------------------------------------------------------

global function tokenize(sequence tableau, atom sep)
-- assembles tokens in a string with separator sep
  sequence s

  if atom(tableau[1]) then
    s = sprintf("%d", tableau[1])
  else
    s = tableau[1]
  end if
  for i = 2 to length(tableau) do
    if atom(tableau[1]) then
      s &= {sep}&sprintf("%d", tableau[i])
    else
      s &= {sep}&tableau[i]
    end if
  end for
--  puts(f_debug, s&"\n")
  return s
end function

------------------------------------------------------------------------------

global function splitFilename(sequence filename)
-- splits a complete filename in path, name and extension
  sequence path, name, ext
  integer l, p, e     -- separators for path and extension
  
  l = length(filename)
  if l = 0 then return {"","",""} end if
  path = {}
  name = {}
  ext = {}
  p = 0
  e = l+1
  for i = l to 1 by -1 do
    if e = l+1 then     -- only one extension allowed, the last one
      if filename[i] = '.' then
        e = i
        ext = filename[e+1..$]
      end if
    end if
    if p = 0 then
      if filename[i] = '\\' then
        p = i
        path = filename[1..p-1]
        exit
      end if
    end if
  end for
  if p = 0 then
    name = filename[1..e-1]
  else
    name = filename[p+1..e-1]
  end if
  return {path,name,ext}
end function

------------------------------------------------------------------------------

global function commonField(sequence sht, integer fld, object val)
-- lists records from sheet sht with value val in field fld
  sequence res

  if fld = 0 then return {} end if
  res = {}
  for i = 1 to length(sht) do
--analyzeSequence(sht[i][fld], "sht[i][fld]", f_debug)
--analyzeSequence(val, "val", f_debug)
    if compare(sht[i][fld], val) = 0 then
      res = append(res, i)
    end if
  end for
  return res
end function

------------------------------------------------------------------------------

global function findID(sequence sht, object id)
-- find the record in sheet sht which ID (value in first column) is id
-- returns the record (sequence) if successful
-- returns 0 (atom) if not
  sequence s
  
  if atom(id) and (id = 0) then return 0 end if
  if sequence(id) and (length(id) = 0) then return 0 end if
  s = commonField(sht, 1, id)
  if length(s) = 1 then
    return sht[s[1]]
  else
    return 0
  end if
end function

------------------------------------------------------------------------------

global function selectedFields(sequence sht, integer fld, object val, sequence fields)
-- list selected fields from sheet sht records with value val in field fld
-- ex: selectedFields(tab, 2, "John", {1,3}) will list items having
-- "John" in second column and return a sequence containing columns 1 and 3
--NB is this still useful? commonField returns the whole records
  sequence s, res
  integer l

  if (fld = 0) or (length(fields) = 0) then return {} end if
  s = commonField(sht, fld, val)
  l = length(s)
  res = {}
  if l = 1 then
    for i = 1 to length(fields) do
      res = append(res, sht[s[1]][fields[i]])
    end for
    return res
  else
    return {}
  end if
end function

------------------------------------------------------------------------------

global function commonFields(sequence sht, sequence fields, sequence vals)
-- list records from sheet sht with selected values in selected fields
-- ex: commonFields(tab, {1,3}, {5,"John"}) will list items having 5 in first column and "John" in third
  sequence res
  integer found
  
  if (length(fields) = 0) or (length(vals) = 0) or (length(fields) != length(vals)) then return {} end if
  res = {}
  for i = 1 to length(sht) do
    found = 1
    for j = 1 to length(fields) do
      if compare(sht[i][fields[j]], vals[j]) != 0 then
        found = 0
      end if
    end for
    if found = 1 then  
      res = append(res, i)
    end if
  end for
  return res
end function

------------------------------------------------------------------------------

global function readCSV(sequence filename)
-- reads a CSV file in a sequence
-- CSV separator is a semicolon
-- CSV strings may be surrounded by quotes
  sequence section
  integer f_in
  object line

  section = {}
  f_in = open(filename, "r")
  if f_in != -1 then
    while 1 do
      line = gets(f_in)
      if atom(line) then
        exit
      end if
      line = line[1..length(line)-1]
      if length(line)>0 then
        section = append(section, strtok(line, ';') )
      end if
    end while
    close(f_in)
  end if
  return section
end function

------------------------------------------------------------------------------

global procedure saveCSV(sequence filename, sequence section)
-- saves a sequence in a CSV file
-- CSV separator is a semicolon
-- CSV strings are surrounded by quotes
  integer f_out, l
  sequence line

  f_out = open(filename, "w")
  if f_out != -1 then
    for i = 1 to length(section) do
      if atom(section[i]) then
        line = object2string(section[i])&"\n"
      else
        l = length(section[i])
        line = {}
        if l > 0 then
          for j = 1 to l-1 do
            if atom(section[i][j]) then
              line &= object2string(section[i][j])&";"
            else
              line &= "\""&object2string(section[i][j])&"\";"
            end if
          end for
          if atom(section[i][l]) then
            line &= object2string(section[i][l])&"\n"
          else
            line &= "\""&object2string(section[i][l])&"\"\n"
          end if
        end if
      end if
      puts(f_out, line)
    end for
    close(f_out)
  end if
end procedure

------------------------------------------------------------------------------

global function readIniFile(sequence filename)
-- reads a Ini file in a sequence { {section name, section content}, ... }
-- each section content is a list of parameters and values { {parameter, value}, ... }
  sequence ItemsList
  sequence Item, s
  object line
  integer f_in

  Item = {}
  s = {}
  ItemsList = {}
  f_in = open(filename, "r")
  if f_in != -1 then
    while 1 do
      line = gets(f_in)
      if atom(line) then
        if length(Item) > 0 then
          Item = append(Item, s)
          ItemsList = append(ItemsList, Item)
        end if
        exit
      end if
      --Add some items to the list
      if line[1] = '[' then
        if length(Item) > 0 then
          Item = append(Item, s)
          s = {}
          ItemsList = append(ItemsList, Item)
          Item = {}
        end if
        line = line[2..length(line)-2]
        Item = append(Item, line)
      else
        line = line[1..length(line)-1]
        if length(line)>0 then
          s = append(s, splitString(line, '=') )
        end if
      end if
    end while
  end if
  close(f_in)
  return ItemsList
end function

------------------------------------------------------------------------------

global function searchINI(sequence ini, sequence section, sequence key)
  sequence s, words

  s = commonField(ini, 1, section)
  analyzeSequence(s, "s", f_debug)
  if length(s) = 1 then
    words = ini[s[1]][2]
    s = commonField(words, 1, key)
    analyzeSequence(s, "s'", f_debug)
    if length(s) = 1 then
      return words[s[1]][2]
    else
      return ""
    end if
  else
    return ""
  end if
end function

------------------------------------------------------------------------------

global function readINI(sequence file, sequence section, sequence name, object value)
  -- reads an element of an INI file
  sequence s
  object x

  s = readIniFile(file)
--analyzeSequence(s, "s", f_debug)
  x = findID(s, section)
--analyzeSequence(x, "x", f_debug)
  if atom(x) then return value end if
  if length(x)<2 then return value end if
  s = x[2]
--analyzeSequence(s, "s", f_debug)
  x = findID(s, name)
--analyzeSequence(x, "x", f_debug)
  if atom(x) then return value end if
  if length(x)<2 then return value end if
  return x[2]
end function

------------------------------------------------------------------------------

global procedure saveIniFile(sequence filename, sequence itemslist)
-- saves a sequence in an Ini file  (see readIni above for sequence format)
  integer f_out

  f_out = open(filename, "w")
  for i = 1 to length(itemslist) do
    puts(f_out, "["&itemslist[i][1]&"]\n")
    for j = 1 to length(itemslist[i][2]) do
      if length(itemslist[i][2]) > 0 then
        puts(f_out, itemslist[i][2][j][1]&"="&itemslist[i][2][j][2]&"\n")
      end if
    end for
    puts(f_out, "\n")
  end for
  close(f_out)
end procedure

------------------------------------------------------------------------------

global procedure writeINI(sequence file, sequence section, sequence name, object value)
  -- writes an element of an INI file
  sequence s, t, ini
  integer sec

  ini = readIniFile(file)
  s = commonField(ini, 1, section)   --recherche section
  if length(s)=0 then    -- nouvelle section
    ini = append(ini, {section, {{name,value}}})
  else   -- section existante
    sec = s[1]
    s = ini[sec]
    if length(s)<2 then  -- section vide
      s = append( s, {{name,value}} )
      ini[sec] = s
    else
      t = commonField(s[2], 1, name)    -- recherche item
      if length(t)=0 then    -- nouvel item
        s[2] = append( s[2], {name,value} )
        ini[sec] = s
      else                   -- item existant, a remplacer
        s[2][t[1]] = {name,value}
        ini[sec] = s
      end if
    end if
  end if
  saveIniFile(file, ini)
end procedure

------------------------------------------------------------------------------

global function round(atom a, integer dec)
-- rounds atom a to n decimals
-- ex: round(112.234, 1) = 112.2
-- n may be negative
-- ex: round(112, -2) = 100
  atom n

  n = power(10, dec)
  return (floor((n*a)+ 0.5) / n)
end function

------------------------------------------------------------------------------

global function insertInSequence(object o, sequence s, integer p)
-- inserts object o in sequence s at position p
-- returns modified sequence
  if (p<1) or (p>length(s)) then return s end if
  if p = 1 then
    return o & s
  else
    return s[1..p-1] & o & s[p..$]
  end if
end function

------------------------------------------------------------------------------

global function removeFromSequence(sequence s, integer p)
-- deletes object at position p in sequence s
  if (p<1) or (p>length(s)) then return s end if
  if p = 1 then
    return s[2..$]
  elsif p = length(s) then
    return s[1..$-1]
  else
    return s[1..p-1] & s[p+1..$]
  end if
end function

------------------------------------------------------------------------------

global procedure hexDump(sequence s, integer output)
-- dumps a sequence in hex form to the specified output
  puts(output, "\n\nAddr  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F\n")
  puts(output, "0000 ")
  for i = 1 to length(s) do
    puts(output, sprintf(" %02x", s[i]))
    if remainder(i, 16) = 0 then
      puts(output, sprintf("\n%04x ", floor(i/16)))
    end if
  end for
  puts(output, "\n\n")
end procedure

------------------------------------------------------------------------------

global function hexString(sequence s)
-- returns hexadecimal representation of a sequence
  sequence result

  result = ""
  for i = 1 to length(s) do
    result &= sprintf("%02x", s[i])
  end for
  return result
end function

------------------------------------------------------------------------------

global function md5sum(sequence filename)
-- returns MD5 checksum of a file
  integer fn
  sequence whole_file, chunk

  whole_file = {}
  fn = open(filename, "rb")  -- an existing file
  while 1 do
    chunk = get_bytes (fn, 100) -- read 100 bytes at a time
    whole_file &= chunk        -- chunk might be empty, that's ok
    if length(chunk) < 100 then exit end if
  end while
  close(fn)
  return hexString(md5(whole_file))
end function

------------------------------------------------------------------------------

global function fileExists(sequence filename)
  if atom(dir(filename)) then
    return 0
  else
    return 1
  end if
end function

------------------------------------------------------------------------------

global function appendIfNew(sequence s, object o)
  if find(o, s) then
    return s
  else
    return append(s, o)
  end if
end function

------------------------------------------------------------------------------

global function utf8_to_ascii(sequence s)
-- created from Tommy's Unicode Library 1.2
-- copyright (c) 2003-2004 Tommy Carlier
  integer i, char, byte, bytes
  sequence res

  i = 1
  res = ""
  while i < length(s) do
    byte = s[i]
    char = 0
    if and_bits(byte,#80) = 0 then bytes = 1
    elsif and_bits(byte,#20) = 0 then bytes = 2
    elsif and_bits(byte,#10) = 0 then bytes = 3
    elsif and_bits(byte,#8) = 0 then bytes = 4
    elsif and_bits(byte,#4) = 0 then bytes = 5
    elsif and_bits(byte,#2) = 0 then bytes = 6
    else return "Incorrect UTF-8 char"
    end if
    byte = and_bits(byte, power(2, 8 - bytes) - 1)
    char += byte * power(64, bytes - 1)
    for n=1 to bytes-1 do
      i += 1
      byte = s[i]
      if byte=0 then
        return "Unexpected end of UTF-8 string"
      elsif byte<0 then
        return "Incorrect UTF-8 char"
      end if
	  byte = and_bits(byte,63)
      char += byte * power(64,bytes-1-n)
    end for
    res &= char
    i += 1
  end while
  return res
end function

